/* audio1: code to interface with SD card
 * Rick Krebs 2025
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdio.h>
#include <zephyr/kernel.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/fs/fs.h>
#include <zephyr/drivers/sdhc.h>
#include <errno.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/sys/printk.h>
#include <zephyr/drivers/spi.h>
#include <zephyr/drivers/pinctrl.h>
#include <zephyr/devicetree/gpio.h>
#include <zephyr/storage/disk_access.h>
#include <ff.h>
#include <stdlib.h>

/* 1000 msec = 1 sec */
#define SLEEP_TIME_MS   500

/* The devicetree node identifier for the "led0" alias. */
#define LED0_NODE DT_ALIAS(led0)
#define DT_DRV_COMPAT xiao_spi 		// nordic_nrf_spim
// #define SPI_DEVICE_NODE DT_NODELABEL(spi0)

#define SD_CS_PIN 28	// chip select pin for SD card (P0.28 on Xiao Sense)
#define SD_CARD_MOUNT_POINT "/SD:"
#define LOG_FILENAME "adc_log.txt"

static FATFS fat_fs;
static struct fs_mount_t mp = {
	.type = FS_FATFS,
	.mnt_point = SD_CARD_MOUNT_POINT,
};
/*
struct spi_cs_control spi_cs = {
	.gpio_dev = DEVICE_DT_GET(DT_NODELABEL(gpioa)),
	.gpio_pin 28,
	.gpio_dt_flags = GPIO_ACTIVE_LOW,
	.delay = 10,
};
*/

static const struct gpio_dt_spec led = GPIO_DT_SPEC_GET(LED0_NODE, gpios);
static const struct device *spi_dev;
static struct spi_cs_control cs_ctrl;

#define SPI_OP SPI_OP_MODE_MASTER | SPI_MODE_CPOL | SPI_MODE_CPHA | SPI_WORD_SET(8) | SPI_LINES_SINGLE
struct spi_dt_spec spispec = SPI_DT_SPEC_GET(DT_NODELABEL(sdhc0), SPI_OP, 0);

/*
spi_dev = device_get_binding(DT_LABEL(DT_DRV_INST(0)));

struct spi_cs_control spi_cs = 
{
	#if DT_NODE_HAS_PROP(DT_DRV_INST(0), cs_gpios)
		.gpio = {
			.pin = DT_GPIO_PIN(DT_DRV_INST(0), cs_gpios),
			.dt_flags = GPIO_ACTIVE_LOW,
		},
	#else
		.gpio = {
			.pin = 0, // Default value or handle error
			.dt_flags = GPIO_ACTIVE_LOW,
		},
	#endif
	.delay = 0,
};

static void spi_init(void)
{
	spi_cs.gpio_dev = device_get_binding(DT_GPIO_LABEL(DT_DRV_INST(0), cs_gpios));

	if (spi_cs.gpio.port == NULL) {
		printk("Could not get gpio device\n");
	} else {
		printk("GPIO device: %s\n", DT_GPIO_LABEL(DT_DRV_INST(0), cs-gpios));
	}

	spi_dev = device_get_binding(DT_LABEL(DT_DRV_INST(0)));

	if (spi_dev == NULL) {
		printk("Could not get %s device\n", DT_LABEL(DT_DRV_INST(0)));
		return;
	} else {
		printk("SPI Device: %s\n", DT_LABEL(DT_DRV_INST(0)));
		printk("SPI CSN %d, MISO %d, MOSI %d, CLK %d\n",
	       DT_GPIO_PIN(DT_DRV_INST(0), cs_gpios),
	       DT_PROP(DT_DRV_INST(0), miso_pin),
	       DT_PROP(DT_DRV_INST(0), mosi_pin),
	       DT_PROP(DT_DRV_INST(0), sck_pin));		
	}
}

void spi_test_send(void)
{
	int err;
	static uint8_t tx_buffer[32];
	static uint8_t rx_buffer[32];

	const struct spi_buf tx_buf = {
		.buf = tx_buffer,
		.len = sizeof(tx_buffer)
	};
	const struct spi_buf_set tx = {
		.buffers = &tx_buf,
		.count = 1
	};

	struct spi_buf rx_buf = {
		.buf = rx_buffer,
		.len = sizeof(rx_buffer),
	};
	const struct spi_buf_set rx = {
		.buffers = &rx_buf,
		.count = 1
	};

	err = spi_transceive(spi_dev, &spi_cfg, &tx, &rx);
	if (err) {
		printk("SPI error: %d\n", err);
	} else {
		// Connect MISO to MOSI for loopback 
		printk("TX sent: %x\n", tx_buffer[0]);
		printk("RX recv: %x\n", rx_buffer[0]);
		tx_buffer[0]++;
	}
}
*/

static int init_sd_card(void)
{
	printk("Inside init_sd_card function...");
	static const char *disk_pdrv = "SD";
	uint64_t memory_size_mb;
	uint32_t block_count;
	uint32_t block_size;
	int err;
 
	err = disk_access_init(disk_pdrv);
	if (err != 0) {
		printk("disk_access_init failed!  Error code: %d (%s) \n", err, strerror(err));
		return -1;
	}

	if (disk_access_ioctl(disk_pdrv, DISK_IOCTL_GET_SECTOR_COUNT, &block_count)) {
		printk("Unable to get sector count!\n");
		return -1;
	}

	if (disk_access_ioctl(disk_pdrv, DISK_IOCTL_GET_SECTOR_SIZE, &block_size)) {
		printk("Unable to get sector size!\n");
		return -1;
	}

	memory_size_mb = (uint64_t) block_count * block_size / (1024 * 1024);
	printk("Memory Size (MB): %u\n", (uint32_t) memory_size_mb);

	mp.fs_data = &fat_fs;
	
	err = fs_mount(&mp);
	if (err) {
		printk("Error mounting fat_fs [%d]\n", err);
		return err;
	}
	
	printk("Disk mounted!\n");

	return 0;
}
/*
static int log_to_sd_card(uint16_t adc_raw, float voltage)
{
	struct fs_file_t file;
	char log_entry[100];
	ssize_t bytes_written;

	fs_file_t_init(&file);

	int err = fs_open(&file, SD_CARD_MOUNT_POINT "/" LOG_FILENAME, FS_O_WRITE | FS_O_CREATE | FS_O_APPEND);

	if (err) {
		printk("Error opening file [%d]\n", err);
		return err;
	}

	snprintf(log_entry, sizeof(log_entry), "ADC raw: %d, Voltage: %.2f V\n", adc_raw, (double) voltage);

	bytes_written = fs_write(&file, log_entry, strlen(log_entry));

	if (bytes_written < 0) {
		printk("Error writing to file [%zd]\n", bytes_written);
		fs_close(&file);

		return (int) bytes_written;
	}

	fs_close(&file);
	printk("Data logged to SD card\n");

	return 0;
}
*/

int main(void)
{
	int ret;
	bool led_state = true;

	printk("Running...\n");

	// initialize SPI device
	spi_dev = device_get_binding("sdhc_spi"); // (SPI_DEVICE_NODE);

	if (!device_is_ready(spi_dev)) {
		printk("SPI device is not ready!\n");
		return -1;
	}
	else {
		printk("SPI initialization successful!\n");
	}

	if (init_sd_card() != 0) {
		printk("Failed to initialize SD card!\n");
		exit (0);
	}

	// configure chip select
	cs_ctrl.gpio.port = DEVICE_DT_GET(DT_NODELABEL(gpio0));
	// cs_ctrl.gpio.pin = ADC_CS_PIN;
	cs_ctrl.gpio.dt_flags = GPIO_ACTIVE_LOW;
	cs_ctrl.delay = 0;

	// fs_open(zipfp, "Maxie.txt", FS_O_RDWR);

	/*
	if (sdhc_card_present(spi_dev))
		printf("hey card is in there!\n");
	else
		printf("error!\n");

	if (!gpio_is_ready_dt(&led)) {
		return 0;
	}
	*/

	ret = gpio_pin_configure_dt(&led, GPIO_OUTPUT_ACTIVE);
	if (ret < 0) {
		return 0;
	}

	while (1) {
		ret = gpio_pin_toggle_dt(&led);
		if (ret < 0) {
			return 0;
		}

		/*
		printk("Attempting to read ADC channel 0\n\r");
		uint16_t adc_raw = read_adc_channel(0);	// read channel 0

		if (adc_raw == 0) {
			printk("ADC read failed or returned 0\n");
		}
		else {
			float voltage = (adc_raw * 3.3f) / 1023.0f;  // assuming 3.3 V reference
			printk("ADC raw value: %d, Voltage: %.2f V\n", adc_raw, voltage);

			if (log_to_sd_card(adc_raw, voltage) != 0) {
				printk("Failed to log data to SD card!\n");
			}
		}
		*/

		printk("\n\r");
		k_msleep(1000);

		led_state = !led_state;
		printf("LED state: %s\n", led_state ? "ON" : "OFF");
		k_msleep(SLEEP_TIME_MS);
	}
	return 0;
}
